import asyncio
from collections import Callable
import concurrent.futures
import inspect

from swim.structures.structs import Absent
from swim.warp.warp import SyncRequest, CommandMessage, Envelope


class ValueDownlink:

    def __init__(self, client):
        self.client = client
        self.host_uri = None
        self.node_uri = None
        self.lane_uri = None

        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

        self.value = None
        self.linked = asyncio.Event(loop=self.client.loop)

    def execute_did_set(self, new_value, old_value):
        # no-op
        pass

    def set_host_uri(self, host_uri):
        self.host_uri = host_uri
        return self

    def set_node_uri(self, node_uri):
        self.node_uri = node_uri
        return self

    def set_lane_uri(self, lane_uri):
        self.lane_uri = lane_uri
        return self

    def did_set(self, function):

        if inspect.iscoroutinefunction(function):
            self.execute_did_set = function
        elif isinstance(function, Callable):
            self.execute_did_set = function
        else:
            raise TypeError('Callback must be a function!')

        return self

    def open(self):
        self.client.schedule_task(self.__open)
        return self

    async def __open(self):
        await self.client.open_websocket(self.host_uri)
        await self.establish_downlink()
        await self.receive_message()

    async def establish_downlink(self):

        sync_request = SyncRequest(self.node_uri, self.lane_uri)
        await self.client.websocket.send(await sync_request.to_recon())

    async def receive_message(self):
        while True:
            message = await self.client.websocket.recv()

            response = await Envelope.parse_recon(message)

            if response.tag == 'linked':
                self.linked.set()
            elif response.tag == 'synced':
                pass
            elif response.tag == 'event':
                await self.set_value(response)

    def set(self, value):
        message = CommandMessage(self.node_uri, self.lane_uri, value)
        self.client.schedule_task(self.send_message, message)

    async def set_value(self, response):
        old_value = self.value

        if response.body == Absent.get_absent():
            self.value = None
        else:
            self.value = response.body.value

        if inspect.iscoroutinefunction(self.execute_did_set):
            self.client.schedule_task(self.execute_did_set, self.value, old_value)
        else:
            self.client.loop.run_in_executor(self.executor, self.execute_did_set, self.value, old_value)

    async def send_message(self, message):
        await self.linked.wait()
        await self.client.websocket.send(await message.to_recon())



from swim.recon.parsers import ReconStructureParser
from swim.recon.writers import ReconStructureWriter


class Recon:
    structure_writer = None
    structure_parser = None

    @staticmethod
    async def parse(recon_string):
        """
        Parse a string and return a Value object.
        :param recon_string:
        :return:
        """
        return await Recon.get_structure_parser().parse_block_string(recon_string)

    @staticmethod
    async def to_string(item):
        """
        Parse an Item object to string.
        :return:
        """
        return await Recon.get_structure_writer().write_item(item)

    @staticmethod
    def get_structure_writer():
        if Recon.structure_writer is None:
            Recon.structure_writer = ReconStructureWriter()

        return Recon.structure_writer

    @staticmethod
    def get_structure_parser():
        if Recon.structure_parser is None:
            Recon.structure_parser = ReconStructureParser()

        return Recon.structure_parser



from abc import ABC, abstractmethod
from enum import Enum


class Item(ABC):

    @property
    @abstractmethod
    def key(self):
        ...

    def concat(self, new_item):

        record = Record.create()
        record.add(self)

        if isinstance(new_item, Record):
            record.add_all(new_item.items)
        else:
            record.add(new_item)

        return record

    @staticmethod
    def from_object(obj):
        if isinstance(obj, Item):
            return obj
        elif isinstance(obj, dict):
            entry = next(iter(obj.items()))
            return Slot.of(entry[0], entry[1])
        else:
            return Value.from_object(obj)

    @staticmethod
    def extant():
        return Extant.get_extant()

    @staticmethod
    def absent():
        return Absent.get_absent()


class Field(Item, ABC):

    @property
    @abstractmethod
    def key(self):
        ...

    @property
    @abstractmethod
    def value(self):
        ...


class Attr(Field):

    def __init__(self, key, value):
        self.__key = key
        self.__value = value

    @property
    def key(self):
        return self.__key

    @property
    def value(self):
        return self.__value

    @staticmethod
    def of(key, value):

        if key is None:
            raise TypeError('key')

        if value is None:
            raise TypeError('value')

        if isinstance(key, Text):
            return Attr(key, value)
        elif isinstance(key, str):
            return Attr(Text.create_from(key), value)
        else:
            raise TypeError('key')

    def key_equals(self, item):
        if isinstance(item, str):
            return self.key.value == item
        elif isinstance(item, Field):
            return self.key == item.key
        else:
            return self.key == item


class Value(Item):

    @property
    def key(self):
        return Value.absent()

    @staticmethod
    def absent():
        return Absent.get_absent()

    def to_value(self):
        return self

    def length(self):
        return 0


class Text(Value):
    empty = None

    def __init__(self, value):
        self.value = value

    @staticmethod
    def create_from(string):
        if not string:
            return Text.get_empty()
        return Text(string)

    @staticmethod
    def get_empty():
        if Text.empty is None:
            Text.empty = Text('')

        return Text.empty

    def string_value(self):
        return self.value

    def float_value(self):
        return float(self.value)


class Num(Value):

    def __init__(self, value):
        self.value = value

    @staticmethod
    def create_from(value):
        return Num(value)

    def num_value(self):
        return self.value


class Bool(Value):
    TRUE = None
    FALSE = None

    def __init__(self, value):
        self.value = value

    @staticmethod
    def get_from(value):

        if value:
            if Bool.TRUE is None:
                Bool.TRUE = Bool(True)

            return Bool.TRUE
        else:
            if Bool.FALSE is None:
                Bool.FALSE = Bool(False)

            return Bool.FALSE


class Absent(Value):
    absent = None

    @staticmethod
    def get_absent():
        if Absent.absent is None:
            Absent.absent = Absent()

        return Absent.absent


class Extant(Value):
    extant = None

    @staticmethod
    def get_extant():
        if Extant.extant is None:
            Extant.extant = Extant()

        return Extant()


class Slot(Field):

    def __init__(self, key, value):
        self.__key = key
        self.__value = value

    @property
    def key(self):
        return self.__key

    @property
    def value(self):
        return self.__value

    @staticmethod
    def of(key, value=None):
        if key is None:
            raise Exception('Key is empty!')

        if value is None:
            value = Value.extant()

        return Slot(key, value)


class Record(Value, ABC):

    @staticmethod
    def create():
        return RecordMap.create()

    @abstractmethod
    def add(self, item):
        ...

    def add_all(self, items):
        changed = False

        for item in items:
            self.add(item)
            changed = True

        return changed

    def slot(self, key, value):

        if isinstance(key, str):
            key = Text.create_from(key)

        if isinstance(value, str):
            value = Text.create_from(value)

        self.add(Slot(key, value))
        return self

    @property
    @abstractmethod
    def size(self):
        ...

    def length(self):
        return self.size

    def get_headers(self, tag):
        head = self.get_head()

        if isinstance(head, Attr) and head.key_equals(tag):
            header = head.value
            if isinstance(header, Record):
                return header
            else:
                return RecordMap.of(header)
        else:
            return None

    def get_head(self):
        return self.get_item(0)

    @abstractmethod
    def get_item(self, index):
        ...

    def bind(self):
        return self


class RecordFlags(Enum):
    IMMUTABLE = 1 << 0
    ALIASED = 1 << 1


class RecordMap(Record):

    def __init__(self, items=None, fields=None, item_count=0, field_count=0, flags=0):
        self.items = items
        self.fields = fields
        self.item_count = item_count
        self.field_count = field_count
        self.flags = flags

    @staticmethod
    def create():
        items = list()
        return RecordMap(items)

    def add(self, item):
        if self.flags & RecordFlags.IMMUTABLE.value:
            raise AttributeError('immutable')
        if self.flags & RecordFlags.ALIASED.value:
            return self.add_aliased(item)
        else:
            return self.add_mutable(item)

    def add_mutable(self, item):
        self.items.append(item)
        self.item_count = self.item_count + 1

        if isinstance(item, Field):
            self.field_count += 1

        return True

    def add_aliased(self, item):
        pass

    @property
    def size(self):
        return self.item_count

    @property
    def tag(self):
        if self.field_count > 0:
            item = self.items[0]
            if isinstance(item, Attr):
                return item.key.value

        return None

    def get_item(self, index):
        if 0 <= index < self.item_count:
            return self.items[index]
        else:
            return Item.absent()

    def get_body(self):
        n = self.item_count

        if n > 2:
            return RecordMapView(self, 1, n).branch()
        elif n == 2:
            item = self.items[1]

            if isinstance(item, Value):
                return item
            else:
                return RecordMap.of(item)

        else:
            return Value.absent()

    @staticmethod
    def of(obj):
        array = list()
        item = Item.from_object(obj)
        array.append(item)

        field_count = 1 if isinstance(item, Field) else 0

        return RecordMap(array, None, 1, field_count, 0)


class RecordMapView(Record):

    def __init__(self, record, lower, upper):
        self.record = record
        self.lower = lower
        self.upper = upper

    def add(self, item):
        pass

    @property
    def size(self):
        return self.upper - self.lower

    def get_item(self, index):
        pass

    def branch(self):
        size = self.size
        fields_count = 0
        copy_index = self.lower
        new_array = list()

        for _ in range(0, size):

            item = self.record.items[copy_index]
            new_array.append(item)

            if isinstance(item, Field):
                fields_count += 1

            copy_index += 1

        return RecordMap(new_array, None, size, fields_count, 0)


class ValueBuilder:

    def __init__(self):
        self.record = None
        self.value = None

    def add(self, item):

        if isinstance(item, Field):
            return self.add_field(item)
        elif isinstance(item, Value):
            return self.add_value(item)
        else:
            raise AssertionError(item)

    def add_field(self, item):
        if self.record is None:
            self.record = Record.create()

            if self.value is not None:
                self.record.add(self.value)
                self.value = None

        self.record.add(item)
        return True

    def add_value(self, item):
        if self.record is not None:
            self.record.add(item)
        elif self.value is None:
            self.value = item
        else:
            self.record = Record.create()
            self.record.add(self.value)
            self.value = None
            self.record.add(item)

        return True

    def bind(self):
        if self.record is not None:
            return self.record
        elif self.value is not None:
            return self.value
        else:
            return Value.absent()



import asyncio
from threading import Thread

import websockets

from swim.client.downlinks.value_downlink import ValueDownlink


class SwimClient:

    def __init__(self):
        self.downlinks = list()
        self.loop = None
        self.websocket = None

    def downlink_value(self):
        return ValueDownlink(self)

    def start(self):
        loop = asyncio.new_event_loop()
        self.loop = loop

        thread = Thread(target=self.__start_event_loop)
        thread.start()

    def stop(self):
        self.schedule_task(self.__stop_client)

    def schedule_task(self, task, *args):

        if len(args) > 0:
            asyncio.run_coroutine_threadsafe(task(*args), loop=self.loop)
        else:
            asyncio.run_coroutine_threadsafe(task(), loop=self.loop)

    def unschedule_task(self, task):
        task.cancel()

    async def __stop_client(self):
        tasks = [task for task in asyncio.all_tasks(self.loop) if task is not asyncio.current_task(self.loop)]
        [task.cancel() for task in tasks]
        await asyncio.gather(*tasks, return_exceptions=True)

        self.loop.stop()

    def __start_event_loop(self):
        asyncio.set_event_loop(self.loop)
        asyncio.get_event_loop().run_forever()

    async def open_websocket(self, host_uri):
        try:
            self.websocket = await websockets.connect(host_uri)
        except Exception as e:
            print(e)



class ReconUtils:

    @staticmethod
    async def is_ident_start_char(c):
        if c:
            c = ord(c)
            return ord('A') <= c <= ord('Z') or c == ord('_') or ord('a') <= c <= ord(
                'z') or 0xc0 <= c <= 0xd6 or 0xd8 <= c <= 0xf6 or 0xf8 <= c <= 0x2ff or 0x370 <= c <= 0x37d or 0x37f <= c <= 0x1fff or 0x200c <= c <= 0x200d or 0x2070 <= c <= 0x218f or 0x2c00 <= c <= 0x2fef or 0x3001 <= c <= 0xd7ff or 0xf900 <= c <= 0xfdcf or 0xfdf0 <= c <= 0xfffd or 0x10000 <= c <= 0xeffff
        else:
            return False

    @staticmethod
    async def is_ident_char(c):
        if c:
            c = ord(c)
            return c == ord('-') or ord('0') <= c <= ord('9') or ord('A') <= c <= ord('Z') or c == ord('_') or ord('a') <= c <= ord(
                'z') or c == 0xb7 or 0xc0 <= c <= 0xd6 or 0xd8 <= c <= 0xf6 or 0xf8 <= c <= 0x37d or 0x37f <= c <= 0x1fff or 0x200c <= c <= 0x200d or 0x203f <= c <= 0x2040 or 0x2070 <= c <= 0x218f or 0x2c00 <= c <= 0x2fef or 0x3001 <= c <= 0xd7ff or 0xf900 <= c <= 0xfdcf or 0xfdf0 <= c <= 0xfffd or 0x10000 <= c <= 0xeffff
        else:
            return False

    @staticmethod
    async def is_space(c):
        if c:
            c = ord(c)
            return c == ord(' ') or c == ord('\t')
        else:
            return False

    @staticmethod
    async def is_digit(c):
        if c:
            c = ord(c)
            return ord('0') <= c <= ord('9')
        else:
            return False



import math
from abc import ABC, abstractmethod

from swim.recon.recon import Recon
from swim.structures.structs import Item, Record, Attr, Value, Num


class Envelope(ABC):

    async def to_recon(self):
        """
        Create a Recon message from Envelope.
        :return:
        """
        return await Recon.to_string(self.to_value())

    def to_value(self):
        """
        Create a value object from Envelope.
        :return:
        """
        return self.get_form().mold(self).to_value()

    @staticmethod
    async def parse_recon(recon_message):
        """
        Parses a Recon message into an Envelope object.
        :return:
        """
        value = await Recon.parse(recon_message)
        return Envelope.create_from_value(value)

    @staticmethod
    def create_from_value(value):
        """
        Create an Envelope from a Value object.
        :param value:
        :return:
        """
        tag = value.tag
        form = Envelope.resolve_form(tag)

        if form is not None:
            return form.cast(value)
        else:
            return None

    @staticmethod
    def resolve_form(tag):
        if tag == 'sync':
            return SyncRequestForm()
        if tag == 'synced':
            return SyncedResponseForm()
        if tag == 'linked':
            return LinkedResponseForm()
        if tag == 'event':
            return EventMessageForm()
        if tag == 'command':
            return CommandMessageForm()

    @abstractmethod
    def get_form(self):
        ...

    @property
    @abstractmethod
    def tag(self):
        ...


class SyncRequest(Envelope):

    def __init__(self, node_uri, lane_uri, prio=0.0, rate=0.0, body=Value.absent()):
        self.node_uri = node_uri
        self.lane_uri = lane_uri
        self.prio = prio
        self.rate = rate
        self.body = body
        self.form = SyncRequestForm()

    @property
    def tag(self):
        return 'sync'

    def get_form(self):
        return self.form


class SyncedResponse(Envelope):

    def __init__(self, node_uri, lane_uri, body=Value.absent()):
        self.node_uri = node_uri
        self.lane_uri = lane_uri
        self.body = body
        self.form = SyncedResponseForm()

    @property
    def tag(self):
        return 'synced'

    def get_form(self):
        return self.form


class LinkedResponse(Envelope):

    def __init__(self, node_uri, lane_uri, prio=0.0, rate=0.0, body=Value.absent()):
        self.node_uri = node_uri
        self.lane_uri = lane_uri
        self.prio = prio
        self.rate = rate
        self.body = body
        self.form = LinkedResponseForm()

    @property
    def tag(self):
        return 'linked'

    def get_form(self):
        return self.form


class CommandMessage(Envelope):

    def __init__(self, node_uri, lane_uri, body=Value.absent()):
        self.node_uri = node_uri
        self.lane_uri = lane_uri
        self.body = body
        self.form = CommandMessageForm()

    @property
    def tag(self):
        return 'command'

    def get_form(self):
        return self.form


class EventMessage(Envelope):

    def __init__(self, node_uri, lane_uri, body=Value.absent()):
        self.node_uri = node_uri
        self.lane_uri = lane_uri
        self.body = body
        self.form = EventMessageForm()

    @property
    def tag(self):
        return 'event'

    def get_form(self):
        return self.form


class LinkAddressedForm(ABC):

    def mold(self, envelope):

        if envelope is not None:

            headers = Record.create().slot('node', envelope.node_uri).slot('lane', envelope.lane_uri)
            prio = envelope.prio

            if prio != 0 and not math.isnan(prio):
                headers.slot('prio', Num.create_from(prio))

            rate = envelope.rate

            if rate != 0 and not math.isnan(rate):
                headers.slot('rate', Num.create_from(rate))

            return Attr.of(self.tag, headers).concat(envelope.body)
        else:
            return Item.extant()

    @property
    @abstractmethod
    def tag(self):
        ...

    def cast(self, item):
        value = item.to_value()
        headers = value.get_headers(self.tag)
        node_uri = None
        lane_uri = None
        prio = 0.0
        rate = 0.0

        for index in range(0, headers.size):
            header = headers.get_item(index)
            key = header.key.string_value()

            if key is not None:

                if key == 'node':
                    node_uri = header.value.string_value()
                elif key == 'lane':
                    lane_uri = header.value.string_value()
                elif key == 'prio':
                    prio = header.value.num_value()
                elif key == 'rate':
                    rate = header.value.num_value()

        if node_uri is not None and lane_uri is not None:
            body = value.get_body()
            return self.create_from(node_uri, lane_uri, prio, rate, body)


class LaneAddressedForm(ABC):

    def mold(self, envelope):

        if envelope is not None:
            headers = Record.create().slot('node', envelope.node_uri).slot('lane', envelope.lane_uri)
            return Attr.of(self.tag, headers).concat(envelope.body)
        else:
            return Item.extant()

    @property
    @abstractmethod
    def tag(self):
        ...

    def cast(self, item):
        value = item.to_value()
        headers = value.get_headers(self.tag)
        node_uri = None
        lane_uri = None

        for index in range(0, headers.size):
            header = headers.get_item(index)
            key = header.key.string_value()

            if key is not None:

                if key == 'node':
                    node_uri = header.value.string_value()
                elif key == 'lane':
                    lane_uri = header.value.string_value()

        if node_uri is not None and lane_uri is not None:
            body = value.get_body()
            return self.create_from(node_uri, lane_uri, body)


class SyncedResponseForm(LaneAddressedForm):

    @property
    def tag(self):
        return 'synced'

    def create_from(self, node_uri, lane_uri, body):
        return SyncedResponse(node_uri, lane_uri, body=body)


class SyncRequestForm(LinkAddressedForm):

    @property
    def tag(self):
        return 'sync'

    def create_from(self, node_uri, lane_uri, prio, rate, body):
        return SyncRequest(node_uri, lane_uri, prio, rate, body=body)


class LinkedResponseForm(LinkAddressedForm):

    @property
    def tag(self):
        return 'linked'

    def create_from(self, node_uri, lane_uri, prio, rate, body):
        return LinkedResponse(node_uri, lane_uri, prio, rate, body=body)


class CommandMessageForm(LaneAddressedForm):

    @property
    def tag(self):
        return 'command'

    def create_from(self, node_uri, lane_uri, body):
        return CommandMessage(node_uri, lane_uri, body=body)


class EventMessageForm(LaneAddressedForm):

    @property
    def tag(self):
        return 'event'

    def create_from(self, node_uri, lane_uri, body):
        return EventMessage(node_uri, lane_uri, body)



from abc import ABC, abstractmethod

from swim.recon.utils import ReconUtils
from swim.structures.structs import ValueBuilder, Text, Bool, Attr, Value, Record, Slot, Num


class Parser(ABC):
    pass


class BlockParser(Parser):

    @staticmethod
    async def parse(message, parser, key_output=None, value_output=None, builder=None):

        char = message.head()

        while await ReconUtils.is_space(char):
            char = message.step()
        # if char == '!' or char == '"' or char == '$' or char == '%' or char == '\'' or char == '(' or char == '+' or char == '-' or char == '@' or c

        if builder is None:
            builder = await parser.create_value_builder()

        if key_output is None:
            key_output = await parser.parse_block_expression(message)

        char = message.head()

        while await ReconUtils.is_space(char):
            char = message.step()

        if message.is_cont():

            if char == ':':
                char = message.step()

            while await ReconUtils.is_space(char):
                char = message.step()

            if value_output is None:
                value_output = await parser.parse_block_expression(message)

            builder.add(await parser.create_slot(key_output, value_output))

            char = message.head()
            if char == ',' or char == ';':
                message.step()
                await BlockParser.parse(message, parser, builder=builder)

            return builder.bind()

        else:
            builder.add(key_output)
            return builder.bind()

    @staticmethod
    async def parse_block(message, parser):
        return await BlockParser.parse(message, parser)


class ReconParser:

    async def parse_attr(self, message):
        return await AttrParser.parse_attr(message, self)

    async def parse_ident(self, message):
        return await IdentParser.parse_ident(message, self)

    async def parse_string(self, message):
        return await StringParser.parse_string(message, self)

    async def parse_number(self, message):
        return await NumberParser.parse_number(message, self)

    async def parse_block_string(self, recon_string):
        message = InputMessage(recon_string)
        return await self.parse_block(message)

    async def parse_block(self, message):
        return await BlockParser.parse_block(message, self)

    async def parse_block_expression(self, message):
        return await self.parse_lambda_func(message)

    async def parse_record(self, message, builder):
        return await RecordParser.parse_record(message, self, builder=builder)

    async def parse_lambda_func(self, message):
        return await LambdaFuncParser.parse_lambda_func(message, self)

    async def parse_conditional_operator(self, message, builder):
        return await ConditionalOperatorParser.parse_conditional_operator(message, self, builder)

    async def parse_or_operator(self, message, builder):
        return await OrOperatorParser.parse_or_operator(message, self, builder)

    async def parse_and_operator(self, message, builder):
        return await AndOperatorParser.parse_and_operator(message, self, builder)

    async def parse_bitwise_or_operator(self, message, builder):
        return await BitwiseOrOperatorParser.parse_bitwise_or_operator(message, self, builder)

    async def parse_bitwise_xor_operator(self, message, builder):
        return await BitwiseXorOperatorParser.parse_bitwise_xor_operator(message, self, builder)

    async def parse_bitwise_and_operator(self, message, builder):
        return await BitwiseAndOperator.parse_bitwise_and_operator(message, self, builder)

    async def parse_comparison_operator(self, message, builder):
        return await ComparisonOperatorParser.parse_comparison_operator(message, self, builder)

    async def parse_attr_expression(self, message, builder):
        return await AttrExpressionParser.parse_attr_expression(message, self, builder)

    async def parse_additive_operator(self, message, builder):
        return await AdditiveOperatorParser.parse_additive_operator(message, self, builder)

    async def parse_multiplicative_operator(self, message, builder):
        return await MultiplicativeOperatorParser.parse_multiplicative_operator(message, self, builder)

    async def parse_prefix_operator(self, message, builder):
        return await PrefixOperatorParser.parse_prefix_operator(message, self, builder)

    async def parse_invoke_operator(self, message, builder):
        return await InvokeOperatorParser.parse_invoke_operator(message, self, builder)

    async def parse_primary(self, message, builder):
        return await PrimaryParser.parse_primary(message, self, builder)

    async def parse_literal(self, message, builder):
        return await LiteralParser.parse_literal(message, self, builder)


class ReconStructureParser(ReconParser):

    async def create_ident(self, value):
        if isinstance(value, Text):
            if value == 'true':
                return Bool.get_from(True)
            elif value == 'false':
                return Bool.get_from(False)

        return Text.create_from(value)

    async def create_attr(self, key, value=Value.extant()):
        return Attr.of(key, value)

    async def create_record_builder(self):
        return Record.create()

    async def create_value_builder(self):
        return ValueBuilder()

    async def create_slot(self, key, value=None):
        return Slot.of(key, value)

    async def create_number(self, value):
        return Num.create_from(value)


class RecordParser(Parser):

    @staticmethod
    async def parse_record(message, parser, builder=None):
        return await RecordParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, key_output=None, value_output=None, builder=None):
        char = message.head()

        if char == '{':
            char = message.step()

        while await ReconUtils.is_space(char):
            char = message.step()

        if key_output is None:
            key_output = await parser.parse_block_expression(message)

        while await ReconUtils.is_space(char):
            char = message.step()

        if message.is_cont():
            if message.head() == ':':
                message.step()

                if value_output is None:
                    value_output = await parser.parse_block_expression(message)

                builder.add(await parser.create_slot(key_output, value_output))

            else:
                builder.add(key_output)

        if message.is_cont():
            char = message.head()

            if char == ',' or char == ';':
                message.step()
                await RecordParser.parse_record(message, parser, builder)

            elif char == '}':
                message.step()
                return builder

class LambdaFuncParser(Parser):

    @staticmethod
    async def parse_lambda_func(message, parser):
        return await LambdaFuncParser.parse(message, parser)

    @staticmethod
    async def parse(message, parser, output=None, builder=None):
        if output is None:
            output = await parser.parse_conditional_operator(message, builder=builder)

        return output


class ConditionalOperatorParser(Parser):

    @staticmethod
    async def parse_conditional_operator(message, parser, builder):
        return await ConditionalOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, if_output=None, then_output=None, else_output=None, builder=None):
        if if_output is None:
            if_output = await parser.parse_or_operator(message, builder)

        return if_output


class OrOperatorParser(Parser):

    @staticmethod
    async def parse_or_operator(message, parser, builder):
        return await OrOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_and_operator(message, builder)

        return lhs_output


class AndOperatorParser(Parser):

    @staticmethod
    async def parse_and_operator(message, parser, builder):
        return await AndOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_bitwise_or_operator(message, builder)

        return lhs_output


class BitwiseOrOperatorParser(Parser):

    @staticmethod
    async def parse_bitwise_or_operator(message, parser, builder):
        return await BitwiseOrOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_bitwise_xor_operator(message, builder)

        return lhs_output


class BitwiseXorOperatorParser(Parser):

    @staticmethod
    async def parse_bitwise_xor_operator(message, parser, builder):
        return await BitwiseXorOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_bitwise_and_operator(message, builder)

        return lhs_output


class BitwiseAndOperator(Parser):

    @staticmethod
    async def parse_bitwise_and_operator(message, parser, builder):
        return await BitwiseAndOperator.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_comparison_operator(message, builder)

        return lhs_output


class ComparisonOperatorParser(Parser):

    @staticmethod
    async def parse_comparison_operator(message, parser, builder):
        return await ComparisonOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, operator_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_attr_expression(message, builder)

        return lhs_output


class AttrExpressionParser(Parser):

    @staticmethod
    async def parse_attr_expression(message, parser, builder):
        return await AttrExpressionParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, field_output=None, value_output=None, builder=None):

        char = message.head()

        while await ReconUtils.is_space(char):
            char = message.step()

        if char == '@':
            if field_output is None:
                field_output = await parser.parse_attr(message)

            if builder is None:
                builder = await parser.create_record_builder()

            builder.add(field_output)

            await AttrExpressionParser.parse(message, parser, builder=builder)

            return builder

        elif await ReconUtils.is_ident_start_char(char) or char == '"' or await ReconUtils.is_digit(char) or char == '-':
            if value_output is None:
                value_output = await parser.parse_additive_operator(message, None)

            if builder is None:
                builder = await parser.create_value_builder()

            builder.add(value_output)
            return builder.bind()

        elif char == '{' or char == '[':
            if builder is None:
                builder = await parser.create_record_builder()

            if value_output is None:
                await parser.parse_additive_operator(message, builder)

            if message.is_cont():
                if message.head() == '@':
                    await AttrExpressionParser.parse(message, parser, builder=builder)


class AttrParser(Parser):

    @staticmethod
    async def parse_attr(message, parser):
        return await AttrParser.parse(message, parser)

    @staticmethod
    async def parse(message, parser, key_output=None, value_output=None):

        char = message.head()
        if char == '@':
            char = message.step()

            if char == '"':
                if key_output is None:
                    key_output = await parser.parse_string(message)
            else:

                if key_output is None:
                    key_output = await parser.parse_ident(message)

                if message.head() == '(' and message.is_cont():
                    message.step()
                else:
                    return await parser.create_attr(key_output)

                if message.head() == ')':
                    message.step()
                    return await parser.create_attr(key_output)
                else:
                    if value_output is None:
                        value_output = await parser.parse_block(message)

                if message.head() == ')':
                    message.step()
                    return await parser.create_attr(key_output, value_output)

        return await parser.create_attr(key_output, value_output)


class IdentParser(Parser):

    @staticmethod
    async def parse_ident(message, parser):
        return await IdentParser.parse(message, parser)

    @staticmethod
    async def parse(message, parser, output=None):

        char = message.head()

        if await ReconUtils.is_ident_start_char(char):
            if output is None:
                output = ''

            output = output + char
            char = message.step()

            while await ReconUtils.is_ident_char(char):
                output = output + char
                char = message.step()

        return await parser.create_ident(output)


class StringParser(Parser):

    @staticmethod
    async def parse_string(message, parser):
        return await StringParser.parse(message, parser)

    @staticmethod
    async def parse(message, parser, output=None):
        char = message.head()

        while await ReconUtils.is_space(char):
            char = message.step()

        if char == '"':

            if output is None:
                output = ''

            char = message.step()

            while char != '"' and message.is_cont():
                output = output + char
                char = message.step()

            message.step()

        return Text.create_from(output)


class NumberParser(Parser):

    @staticmethod
    async def parse_number(message, parser):
        return await NumberParser.parse(message, parser)

    @staticmethod
    async def parse(message, parser, value_output=None, sign_output=1):

        char = message.head()

        if char == '-':
            sign_output = -1
            char = message.step()

        if char == '0':
            char = message.step()
        elif '1' <= char <= '9':
            value_output = sign_output * int(char)
            char = message.step()

            while message.is_cont() and await ReconUtils.is_digit(char):
                value_output = 10 * value_output + sign_output * int(char)
                char = message.step()

        if message.is_cont():
            if char == '.':
                return await DecimalParser.parse_decimal(message, parser, value_output, sign_output)
            else:
                return await parser.create_number(value_output)
        else:
            return await parser.create_number(value_output)


class DecimalParser(Parser):

    @staticmethod
    async def parse_decimal(message, parser, value_output=None, sign_output=None):
        builder = ''

        if sign_output < 0 and value_output is None:
            builder += '-0'
        else:
            if value_output is None:
                value_output = 0

            builder += str(value_output)

        return await DecimalParser.parse(message, parser, builder)

    @staticmethod
    async def parse(message, parser, builder):

        char = message.head()

        if char == '.':
            builder += '.'
            message.step()

            if message.is_cont():
                char = message.head()

                if await ReconUtils.is_digit(char):
                    builder += char
                    char = message.step()

                while message.is_cont() and await ReconUtils.is_digit(char):
                    builder += char
                    char = message.step()

                return await parser.create_number(float(builder))


class AdditiveOperatorParser(Parser):

    @staticmethod
    async def parse_additive_operator(message, parser, builder):
        return await AdditiveOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, operator_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_multiplicative_operator(message, builder)

            char = message.head()

            if char == '+':
                pass
            else:
                return lhs_output


class MultiplicativeOperatorParser(Parser):

    @staticmethod
    async def parse_multiplicative_operator(message, parser, builder):
        return await MultiplicativeOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, lhs_output=None, operator_output=None, rhs_output=None, builder=None):
        if lhs_output is None:
            lhs_output = await parser.parse_prefix_operator(message, builder)

            char = message.head()

            if char == '*':
                pass
            elif char == '/':
                pass
            elif char == '%':
                pass
            else:
                return lhs_output


class PrefixOperatorParser(Parser):

    @staticmethod
    async def parse_prefix_operator(message, parser, builder):
        return await PrefixOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, operator_output=None, rhs_output=None, builder=None):

        char = message.head()

        if char == '!':
            pass
        elif char == '~':
            pass
        elif char == '+':
            pass
        else:
            return await parser.parse_invoke_operator(message, builder)


class InvokeOperatorParser(Parser):

    @staticmethod
    async def parse_invoke_operator(message, parser, builder):
        return await InvokeOperatorParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, expr_output=None, args_output=None, builder=None):
        if expr_output is None:
            expr_output = await parser.parse_primary(message, builder)

            char = message.head()

            if char == '(':
                pass
            else:
                return expr_output


class PrimaryParser(Parser):

    @staticmethod
    async def parse_primary(message, parser, builder):
        return await PrimaryParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, expr_output=None, builder=None):

        char = message.head()

        if char == '(':
            pass
        else:
            if expr_output is None:
                expr_output = await parser.parse_literal(message, builder)

            return expr_output


class LiteralParser(Parser):

    @staticmethod
    async def parse_literal(message, parser, builder):
        return await LiteralParser.parse(message, parser, builder=builder)

    @staticmethod
    async def parse(message, parser, value_output=None, builder=None):
        char = message.head()

        if char == '(':
            pass
        elif char == '{':
            if builder is None:
                builder = await parser.create_record_builder()

            await parser.parse_record(message, builder)

        elif char == '[':
            pass
        elif await ReconUtils.is_ident_start_char(char):
            value_output = await parser.parse_ident(message)
        elif char == '"':
            value_output = await parser.parse_string(message)
        elif char == '-' or await ReconUtils.is_digit(char):
            value_output = await parser.parse_number(message)
        if builder is None:
            builder = await parser.create_value_builder()

        if value_output is not None:
            builder.add(value_output)

        return builder.bind()


class InputMessage:

    def __init__(self, message):
        self.message = message
        self.index = 0

    def head(self):

        if self.is_cont():
            return self.message[self.index]
        else:
            return ''

    def step(self):
        self.index = self.index + 1
        return self.head()

    def strip(self):
        self.message = self.message.strip()
        return self

    def is_cont(self):
        if self.index >= len(self.message):
            return False
        else:
            return True



from abc import ABC, abstractmethod

from swim.recon.utils import ReconUtils
from swim.structures.structs import Field, Attr, Slot, Value, Record, Text, Absent, Num, Extant


class Writer(ABC):
    ...


class BlockWriter(Writer):

    @staticmethod
    async def write_item(item, writer):
        if isinstance(item, Field):
            return await writer.write_slot(await writer.key(item), await writer.value(item))
        else:
            return await writer.write_item(item)

    @staticmethod
    async def write_block(items, writer):
        return await BlockWriter.write(items, writer, first=True)

    @staticmethod
    async def write(items, writer, first, in_braces=False):
        output = ''

        for item in items:

            if isinstance(item, Attr):
                item_text = await writer.write_item(item)
            elif isinstance(item, Value) and not isinstance(item, Record):
                item_text = await writer.write_item(item)
            else:

                if not first:
                    output += ','
                elif isinstance(item, Slot):
                    if len(output) > 0 and output[-1] != '(':
                        output += '{'
                        in_braces = True

                item_text = await writer.write_item(item)

                first = False

            if item_text:
                output += item_text

        if in_braces:
            output += '}'

        return output


class ReconWriter(ABC):

    @abstractmethod
    async def write_value(self, value):
        ...

    @abstractmethod
    async def key(self, item):
        ...

    @abstractmethod
    async def value(self, item):
        ...

    async def is_ident(self, value):

        if len(value) == 0:
            return False

        if not await ReconUtils.is_ident_start_char(value[0]):
            return False

        for char in value:
            if not await ReconUtils.is_ident_char(char):
                return False

        return True

    async def write_attr(self, key, value):
        return await AttrWriter.write_attr(key, value, self)

    async def write_slot(self, key, value):
        return await SlotWriter.write_slot(key, value, self)

    async def write_text(self, value):
        if await self.is_ident(value):
            return await IdentWriter.write_ident(value)
        else:
            return await StringWriter.write_string(value)

    async def write_number(self, value):
        return await NumberWriter.write_number(value)

    async def write_absent(self):
        pass

    async def write_record(self, record):
        if record.size > 0:
            return await BlockWriter.write_block(record.items, ReconStructureWriter())
        else:
            return '{}'


class ReconStructureWriter(ReconWriter):

    async def write_item(self, item):

        if isinstance(item, Field):
            if isinstance(item, Attr):
                return await self.write_attr(item.key, item.value)
            elif isinstance(item, Slot):
                return await self.write_slot(item.key, item.value)
        elif isinstance(item, Value):
            return await self.write_value(item)

        raise AttributeError(f'No Recon serialization for {item}')

    async def write_value(self, value):
        if isinstance(value, Record):
            return await self.write_record(value)
        elif isinstance(value, Text):
            return await self.write_text(value.string_value())
        elif isinstance(value, Num):
            return await self.write_number(value.num_value())
        elif isinstance(value, Absent):
            return await self.write_absent()

    async def key(self, item):
        return item.key

    async def value(self, item):
        return item.value


class AttrWriter(Writer):

    @staticmethod
    async def write_attr(key, value, writer):

        output = '@'

        key_text = await writer.write_value(key)

        if key_text:
            output += key_text

        if not isinstance(value, Extant):

            output += '('

            value_text = await writer.write_value(value)

            if value_text:
                output += value_text

            output += ')'

        return output


class SlotWriter(Writer):

    @staticmethod
    async def write_slot(key, value, writer):
        output = ''

        key_text = await writer.write_value(key)

        if key_text:
            output += key_text

        output += ':'

        value_text = await writer.write_value(value)

        if value_text:
            output += value_text

        return output


class StringWriter(Writer):

    @staticmethod
    async def write_string(value):
        output = '"'

        if value:
            output += value

        output += '"'

        return output


class NumberWriter(Writer):

    @staticmethod
    async def write_number(value):
        output = ''

        if value:
            output += str(value)

        return output


class IdentWriter(Writer):

    @staticmethod
    async def write_ident(value):
        output = ''

        if value:
            output += value

        return output



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='swimai',
    version='0.0.1.dev2',
    author='Dobromir Marinov',
    author_email='dobromir@swim.it',
    description='Standalone Python framework for building massively real-time streaming WARP clients.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/DobromirM/swim-system-python',
    packages=setuptools.find_packages(exclude=['test']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    keywords='swim client',
    install_requires=['websockets==8.0.2'],
)



import asyncio
import unittest

from aiounittest import async_test
from swim.structures.structs import Num, Text, RecordMap, Attr, Slot
from swim.warp.warp import Envelope, SyncedResponse, CommandMessage, SyncRequest, LinkedResponse, EventMessage


class TestParser(unittest.TestCase):

    @async_test
    async def test_parse_sync(self):
        # Given
        message = '@sync(node: foo, lane: bar)'
        expected = SyncRequest('foo', 'bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_sync_escaped(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")'
        expected = SyncRequest('bar/baz/2', 'foo/bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_sync_body_int(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")43'
        expected = SyncRequest('bar/baz/2', 'foo/bar', body=Num.create_from(43))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_sync_body_float(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")0.31'
        expected = SyncRequest('bar/baz/2', 'foo/bar', body=Num.create_from(0.31))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_sync_body_string(self):
        # Given
        message = '@sync(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        expected = SyncRequest('bar/baz/2', 'foo/bar', body=Text.create_from('Hello, World'))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_sync_prio(self):
        # Given
        message = '@sync(node: foo, lane: bar, prio: 3.2)'
        expected = SyncRequest('foo', 'bar', prio=3.2)
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.prio, actual.prio)

    @async_test
    async def test_parse_sync_rate(self):
        # Given
        message = '@sync(node: foo, lane: bar, rate: 33)'
        expected = SyncRequest('foo', 'bar', rate=33)
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.rate, actual.rate)

    @async_test
    async def test_parse_sync_prio_rate(self):
        # Given
        message = '@sync(node: foo, lane: bar, prio: 13, rate: 37)'
        expected = SyncRequest('foo', 'bar', prio=13, rate=37)
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.prio, actual.prio)
        self.assertEqual(expected.rate, actual.rate)

    @async_test
    async def test_parse_synced(self):
        # Given
        message = '@synced(node: foo, lane: bar)'
        expected = SyncedResponse('foo', 'bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_synced_escaped(self):
        # Given
        message = '@synced(node: "foo/bar", lane: "lane/uri/test")'
        expected = SyncedResponse('foo/bar', 'lane/uri/test')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_synced_body_int(self):
        # Given
        message = '@synced(node: foo, lane: bar)33'
        expected = SyncedResponse('foo', 'bar', body=Num.create_from(33))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_synced_body_float(self):
        # Given
        message = '@synced(node: foo, lane: bar)37.13'
        expected = SyncedResponse('foo', 'bar', body=Num.create_from(37.13))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_synced_body_string(self):
        # Given
        message = '@synced(node: foo, lane: bar)"Hello, World"'
        expected = SyncedResponse('foo', 'bar', body=Text.create_from('Hello, World'))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_linked(self):
        # Given
        message = '@linked(node: foo, lane: bar)'
        expected = LinkedResponse('foo', 'bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_linked_escaped(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")'
        expected = LinkedResponse('bar/baz/2', 'foo/bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_linked_body_int(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")9999999'
        expected = LinkedResponse('bar/baz/2', 'foo/bar', body=Num.create_from(9999999))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_linked_body_float(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")-0.00031'
        expected = LinkedResponse('bar/baz/2', 'foo/bar', body=Num.create_from(-0.00031))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_linked_body_string(self):
        # Given
        message = '@linked(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        expected = LinkedResponse('bar/baz/2', 'foo/bar', body=Text.create_from('Hello, World'))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_linked_prio(self):
        # Given
        message = '@linked(node: foo, lane: bar, prio: 1000.3)'
        expected = LinkedResponse('foo', 'bar', prio=1000.3)
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.prio, actual.prio)

    @async_test
    async def test_parse_linked_rate(self):
        # Given
        message = '@linked(node: "foo", lane: "1/bar/", rate: 33)'
        expected = LinkedResponse('foo', '1/bar/', rate=33)
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.rate, actual.rate)

    @async_test
    async def test_parse_linked_prio_rate(self):
        # Given
        message = '@linked(node: foo, lane: bar, prio: 13, rate: 37)'
        expected = LinkedResponse('foo', 'bar', prio=13, rate=37)
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.prio, actual.prio)
        self.assertEqual(expected.rate, actual.rate)

    @async_test
    async def test_parse_command(self):
        # Given
        message = '@command(node: foo, lane: bar)'
        expected = CommandMessage('foo', 'bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_command_escaped(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")'
        expected = CommandMessage('foo/bar', 'lane/uri/test')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_command_body_int(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")-24'
        expected = CommandMessage('foo/bar', 'lane/uri/test', Num.create_from(-24))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_command_body_float(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")-0.5'
        expected = CommandMessage('foo/bar', 'lane/uri/test', Num.create_from(-0.5))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_command_body_string(self):
        # Given
        message = '@command(node: "foo/bar", lane: "lane/uri/test")"0.32"'
        expected = CommandMessage('foo/bar', 'lane/uri/test', Text.create_from('0.32'))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_command_body_remove(self):
        # Given
        message = '@command(node:"/unit/foo",lane:shoppingCart)@remove(key:FromClientLink)'
        expected = CommandMessage('/unit/foo', 'shoppingCart',
                                  body=RecordMap.of(
                                      Attr.of(Text.create_from('remove'),
                                              RecordMap.of(Slot.of(Text.create_from('key'), Text.create_from('FromClientLink'))))))

        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.items[0].key.value, actual.body.items[0].key.value)
        self.assertEqual(expected.body.items[0].value.items[0].key.value, actual.body.items[0].value.items[0].key.value)
        self.assertEqual(expected.body.items[0].value.items[0].value.value, actual.body.items[0].value.items[0].value.value)

    @async_test
    async def test_parse_event(self):
        # Given
        message = '@event(node: foo, lane: bar)'
        expected = EventMessage('foo', 'bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_event_escaped(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")'
        expected = EventMessage('bar/baz/2', 'foo/bar')
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)

    @async_test
    async def test_parse_event_body_int(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")332'
        expected = EventMessage('bar/baz/2', 'foo/bar', body=Num.create_from(332))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_event_body_float(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")0.1'
        expected = EventMessage('bar/baz/2', 'foo/bar', body=Num.create_from(0.1))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_event_body_string(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")"Hello, World"'
        expected = EventMessage('bar/baz/2', 'foo/bar', body=Text.create_from('Hello, World'))
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.node_uri, actual.node_uri)
        self.assertEqual(expected.lane_uri, actual.lane_uri)
        self.assertEqual(expected.body.value, actual.body.value)

    @async_test
    async def test_parse_event_body_object(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")@Person{name:Bar,age:14,salary:5.9}'
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual('event', actual.tag)
        self.assertEqual('bar/baz/2', actual.node_uri)
        self.assertEqual('foo/bar', actual.lane_uri)

        self.assertEqual(4, len(actual.body.items))
        self.assertEqual('Person', actual.body.items[0].key.value)
        self.assertEqual('name', actual.body.items[1].key.value)
        self.assertEqual('Bar', actual.body.items[1].value.value)
        self.assertEqual('age', actual.body.items[2].key.value)
        self.assertEqual(14, actual.body.items[2].value.value)
        self.assertEqual('salary', actual.body.items[3].key.value)
        self.assertEqual(5.9, actual.body.items[3].value.value)

    @async_test
    async def test_parse_event_body_nested(self):
        # Given
        message = '@event(node: "bar/baz/2", lane: "foo/bar")@Person{name:Bar,age:14,salary:5.9,friend:@Person{name:Foo,age:18,salary:99.9}'
        # When
        requests = await asyncio.gather(Envelope.parse_recon(message))
        actual = requests[0]
        # Then
        self.assertEqual('event', actual.tag)
        self.assertEqual('bar/baz/2', actual.node_uri)
        self.assertEqual('foo/bar', actual.lane_uri)

        self.assertEqual(5, len(actual.body.items))
        self.assertEqual('Person', actual.body.items[0].key.value)
        self.assertEqual('name', actual.body.items[1].key.value)
        self.assertEqual('Bar', actual.body.items[1].value.value)
        self.assertEqual('age', actual.body.items[2].key.value)
        self.assertEqual(14, actual.body.items[2].value.value)
        self.assertEqual('salary', actual.body.items[3].key.value)
        self.assertEqual(5.9, actual.body.items[3].value.value)
        self.assertEqual('friend', actual.body.items[4].key.value)
        self.assertEqual(4, len(actual.body.items[4].value.items))
        self.assertEqual('Person', actual.body.items[4].value.items[0].key.value)
        self.assertEqual('name', actual.body.items[4].value.items[1].key.value)
        self.assertEqual('Foo', actual.body.items[4].value.items[1].value.value)
        self.assertEqual('age', actual.body.items[4].value.items[2].key.value)
        self.assertEqual(18, actual.body.items[4].value.items[2].value.value)
        self.assertEqual('salary', actual.body.items[4].value.items[3].key.value)
        self.assertEqual(99.9, actual.body.items[4].value.items[3].value.value)
